/*
  Copyright 2005, 2006 Damien Stehlé.

  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the
  Free Software Foundation; either version 2 of the License, or (at your
  option) any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
  more details.

  You should have received a copy of the GNU General Public License along
  with this program; see the file COPYING.  If not, write to the Free
  Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
  02111-1307, USA.

  This program implements ideas from the paper 
  "Searching Worst Cases of a One-Variable Function Using Lattice Reduction",
  by Damien Stehlé, Vincent Lefèvre and Paul Zimmermann, in 
  IEEE Transactions on Computers (March 2005).

  The algorithms described in the afore-mentionned paper were improved further
  in Damien Stehlé's PhD thesis, available at:
  http://www.loria.fr/~stehle/THESE.html
  The present code implements the algorithm described in the thesis.
*/

#define VERSION "1.0"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "gmp.h"
#include "mpfr.h"
#include "intclass.h"
#include "mat.h"
#include "bacsel.h"

/* Choose the variant of lll that will be used */
#ifdef PROVED 
#include "fpclass.h"
#include "basic.h"
#include "proved.h"
#else
#include "fast.h"
#endif

// #define VBSE
// #define VBSE1
// #define DEB
// #define DEBDEB
// #define DEB2
// #define VERIF
// #define AUTOMATIC

#ifdef AUTOMATIC
#define FAILURE_LOWERBOUND 1.5
#define FAILURE_UPPERBOUND 7
#endif

/* Choose your basis. Default: 10. */
#ifndef BASIS
#define BASIS 10
#endif


/* Choose your function. Default: exp. */
#define EXPO       1
#define TWOTOX     2
#define LOG2       3
#define LOG10      4
#define LN         5  
#define SIN        6
#define COS        7
#define ONEOVERX   9
#define HALL       10   /* x -> x^(3/2), related to Hall's conjecture. */
#define SQROOT     11
#define SQRTINVERT 12

#ifndef FUNCTION
#define FUNCTION EXPO
#endif

mpfr_t target;
int near;

#define nprimes 10
int primes[] = {3, 5, 7, 11, 13, 17, 19, 23, 29, 31}; 

/*********************************************************************/
/**********************************************************************
BASIC STUFF ON POLYNOMIALS OF THE SHAPE 
1 x ... x^ad y yx ... yx^((a-1)d) .... y^a

Rank= dx+dy/2 * (d(2a-dy+1)+2)   [if we start at rank=0].

 dx+(dy*(d*(2*alpha-dy+1)+2))/2+1
**********************************************************************/
/*********************************************************************/


/* I/O of polynomials */

void
print_pol(int *P, int d, int alpha) 
{
  int k = 0, xd, yd; 

  for (yd = 0; yd <= alpha; yd++) 
    for (xd = 0; xd <= d*alpha - d*yd; xd++) 
      {
	if (P[k] != 0)
	  printf("%d*X^%d*Y^%d + ", P[k], xd, yd);  
	k++; 
      }
}

void
print_pol_mpz(mpz_t *P, int d, int alpha) 
{
  int k = 0, xd, yd; 

  for (yd = 0; yd <= alpha; yd++) 
    for (xd = 0; xd <= d*alpha - d*yd; xd++) 
      {
	gmp_printf("%Zd*X^%d*Y^%d + ", P[k], xd, yd);  
	k++; 
      }
}

/* Derivatives of polynomials: diff1 is toward x, diff2 is toward y,
   diff is mod p, deriv_pol is over the mpz_t's */

void 
diff1(int *p1x, int *P1, int p, int d, int alpha) 
{
  int k = 0, xd, yd; 

  for (yd = 0; yd <= alpha; yd++) 
    for (xd = 0; xd <= d*alpha - d*yd; xd++) 
      {
	if (xd < d*alpha - d*yd) 
	  p1x[k] = (P1[k+1] * (xd + 1)) % p; 
	else p1x[k] = 0; 
	k++; 
      }
}

void 
diff2(int *p1y, int *P1, int p, int d, int alpha) 
{
  int k = 0, xd, yd; 

  for (yd = 0; yd <= alpha; yd++) 
    for (xd = 0; xd <= d*alpha - d*yd; xd++) 
      {
	if (xd <= d*alpha - d*yd - d) 
	  p1y[k] = (P1[k + d*alpha - d*yd + 1] * (yd + 1)) % p; 
	else p1y[k] = 0; 
	k++; 
      }
}

void 
deriv_pol1(mpz_t *p1x, mpz_t *P1, int d, int alpha)
{
  int k = 0, xd, yd; 

  for (yd = 0; yd <= alpha; yd++) 
    for (xd = 0; xd <= d*(alpha - yd); xd++) 
      {
	if (xd < d*(alpha - yd)) 
	  mpz_mul_ui(p1x[k], P1[k+1], (xd + 1)); 
	else mpz_set_ui(p1x[k], 0); 
	k++; 
      }
}

void 
deriv_pol2(mpz_t *p1y, mpz_t *P1, int d, int alpha)
{
  int k = 0, xd, yd; 

  for (yd = 0; yd <= alpha; yd++) 
    for (xd = 0; xd <= d*(alpha - yd); xd++) 
      {
	if (xd <= d*(alpha - yd) - d) 
	  mpz_mul_ui(p1y[k], P1[k + d*alpha - d*yd + 1], (yd + 1)); 
	else mpz_set_ui(p1y[k], 0); 
	k++; 
      }
}


/* Evaluation of polynomials: mod p, over the mpz_t's */

int 
eval_mod(int *P1, int x, int y, int p, int d, int alpha)
{
  int *xpows = malloc ((d*alpha + 1) * sizeof(int)); 
  int *ypows = malloc ((alpha + 1) * sizeof(int)); 
  int k = 0, j, xd, yd, res = 0; 
  
  xpows [0] = 1; 
  for (j = 1; j <= d*alpha; j++)  
    xpows[j] = (xpows[j-1] * x) % p; 
  
  ypows[0] = 1; 
  for (j = 1; j <= alpha; j++) 
    ypows[j] = (ypows[j-1] * y) % p; 

  for (yd = 0; yd <= alpha; yd++) 
    for (xd = 0; xd <= d*alpha - d*yd; xd++) 
      {
	res = (res + P1[k] * xpows[xd] * ypows[yd]) % p; 
	k++; 
      }

  free (xpows);
  free (ypows);

  return res; 
}

void
eval_pol(mpz_t res, mpz_t *P1, mpz_t x, mpz_t y, mpz_t M, int d, int alpha)
{
  mpz_t *xpows = malloc ((d*alpha + 1) * sizeof(mpz_t)); 
  mpz_t *ypows = malloc ((alpha + 1) * sizeof(mpz_t)); 
  int k = 0, j, xd, yd; 
  mpz_t tmp;

  mpz_init (tmp); 
  mpz_init_set_ui (xpows[0], 1); 

  for (j = 1; j <= d*alpha; j++) { 
    mpz_mul (tmp, xpows[j-1], x); 
    mpz_init (xpows[j]); 
    mpz_fdiv_r (xpows[j], tmp, M); 
  }

  mpz_init_set_ui (ypows [0], 1); 
  for (j = 1; j <= alpha; j++) { 
    mpz_mul (tmp, ypows[j-1], y); 
    mpz_init (ypows[j]); 
    mpz_fdiv_r (ypows[j], tmp, M); 
  }

  mpz_set_ui (res, 0); 
  
  for (yd = 0; yd <= alpha; yd++) 
    for (xd = 0; xd <= d*(alpha - yd); xd++) 
      {
	mpz_mul (tmp, xpows[xd], ypows[yd]); 
	mpz_fdiv_r (tmp, tmp, M); 
	mpz_mul (tmp, tmp, P1[k]); 
	mpz_fdiv_r (tmp, tmp, M); 
	
	mpz_add (res, res, tmp); 
	k++; 
      }
  mpz_fdiv_r(res, res, M); 

  mpz_clear (tmp);
  for (j=0; j<= d*alpha; j++) mpz_clear (xpows[j]);
  free (xpows);
  for (j=0; j<= alpha; j++) mpz_clear (ypows[j]);
  free (ypows);
}

/*********************************************************************/
/**********************************************************************
    HENSEL lifting for a set of 2 polynomials.
    --> to find the small common roots of 2 bivariate polynomials 
        over the integers
**********************************************************************/
/*********************************************************************/


/* Finds a good p in the list, ie, either no root mod p, or an invertible
   Jacobian mod p for all root,
   then returns all the roots mod p, set nb to the number of roots,
   set P to p
*/

int **
roots_mod_p (mpz_t *P1, mpz_t *P2, int *P, int *nb, int d, int alpha)
{ 
  int ncoefs = (alpha + 1)*(d*alpha + 2)/2;
  int *p1 = malloc (ncoefs * sizeof(int)); 
  int *p2 = malloc (ncoefs * sizeof(int)); 
  int *p1x = malloc (ncoefs * sizeof(int)); 
  int *p2x = malloc (ncoefs * sizeof(int)); 
  int *p1y = malloc (ncoefs * sizeof(int));
  int *p2y = malloc (ncoefs * sizeof(int)); 
  int i, **z=NULL, flag=0, j, p=0, x, y, zz1=0, zz2=0, zz3=0, zz4=0; 
  mpz_t toto; 

  mpz_init (toto); 

  /* try to find a good p: either no solution mod p, or an invertible
     Jacobian matrix mod p 
  */
  
  i = 0 ;
  while ((i<nprimes) && (flag ==0))
    { 
      flag = 1; 
      p = primes[i]; 

      z = (int **) malloc(p*p*sizeof(int *)); 
      for (j = 0; j < p*p; j++) 
	z[j] = (int *) malloc (2 * sizeof(int)); 

#ifdef DEBDEB
      printf("** p = %d\n", p); 
#endif

      /* p1 is P1 mod p, p2 is P2 mod p */

      for (j = 0; j < ncoefs; j++) { 
	p1[j] = mpz_fdiv_ui (P1[j], p); 
	p2[j] = mpz_fdiv_ui (P2[j], p); 
      }

      /*Compute the Jacobian matrix mod p */
      diff1(p1x, p1, p, d, alpha); 
      diff2(p1y, p1, p, d, alpha); 
      diff1(p2x, p2, p, d, alpha); 
      diff2(p2y, p2, p, d, alpha);

#ifdef DEBDEB      
      printf ("p1 = "); 
      print_pol (p1, d, alpha); 
      printf ("\np2 = "); 
      print_pol (p2, d, alpha); 
      printf ("\np1x = "); 
      print_pol (p1x, d, alpha); 
      printf ("\np1y = "); 
      print_pol (p1y, d, alpha); 
      printf ("\np2x = "); 
      print_pol (p2x, d, alpha); 
      printf ("\np2y = "); 
      print_pol (p2y, d, alpha);
      printf ("\n");
#endif      

      *nb = 0; 

      for (x = 0; flag && (x < p); x++) 
	for (y = 0; flag && (y < p); y++) { 

#ifdef DEBDEB	  
	  printf("x = %d y = %d P1(x, y) = %d\n", 
		 x, y, eval_mod(p1, x, y, p, d, alpha));
	  eval_mod(p2, x, y, p, d, alpha);
	  printf("x = %d y = %d P2(x, y) = %d\n", 
		 x, y, eval_mod(p2, x, y, p, d, alpha));
#endif
	  
	  if (!eval_mod(p1, x, y, p, d, alpha) 
	      && !eval_mod(p2, x, y, p, d, alpha))
	    { 
#ifdef DEBDEB      
	      printf("x = %d y = %d\n", x, y); 
#endif

	      zz1 = eval_mod (p1x, x, y, p, d, alpha); 
	      zz2 = eval_mod (p1y, x, y, p, d, alpha); 
	      zz3 = eval_mod (p2x, x, y, p, d, alpha); 
	      zz4 = eval_mod (p2y, x, y, p, d, alpha); 
	
#ifdef DEBDEB      
	      printf("zz1 = %d zz2 = %d zz3 = %d zz4 = %d\n", 
		     zz1, zz2, zz3, zz4); 
#endif
	      if ((zz1*zz4 - zz2*zz3) % p == 0) flag = 0;
	      
	      z[*nb][0] = x; 
	      z[*nb][1] = y; 
	      (*nb)++; 
	    }
	}
      if ((*nb==0)||(flag==0)){
	for (j = 0; j < p*p ; j++) free(z[j]); 
	free(z); 
	*nb = 0;
      }
      i++;
    }

  if (flag == 1) *P = p;
  else *P = 0;

  free(p1x); 
  free(p1y); 
  free(p2x); 
  free(p2y); 
  free(p1); 
  free(p2); 
  mpz_clear (toto);
  return z;
}

/* Finds the roots (x,y) of P1 and P2 over the integers 
by using Hensel's lifting with a good p. 
Store the roots in Sol, and their number in Nb
*/

int 
hensel(mpz_t **Sol, int *Nb, 
       mpz_t *P1, mpz_t *P2, mpz_t bnd, int d, int alpha) 
{ 
  int ncoefs = (alpha + 1)*(d*alpha + 2)/2;
  int p, nb, j, success; 
  mpz_t tmp1, tmp2, tmp3, X1, Y1, D1X, D2X, D1Y, D2Y, X, Y, M, bnd1; 
  int **r; 
  mpz_t *P1X, *P2X, *P1Y, *P2Y; 
  P1X = (mpz_t *) malloc (ncoefs * sizeof(mpz_t)); 
  P2X = (mpz_t *) malloc (ncoefs * sizeof(mpz_t)); 
  P1Y = (mpz_t *) malloc (ncoefs * sizeof(mpz_t)); 
  P2Y = (mpz_t *) malloc (ncoefs * sizeof(mpz_t)); 

  for (j=0; j<=ncoefs-1; j++)
    {
      mpz_init (P1X[j]);
      mpz_init (P2X[j]);
      mpz_init (P1Y[j]);
      mpz_init (P2Y[j]);
    }
  
  mpz_init(tmp1); 
  mpz_init(tmp2); 
  mpz_init(tmp3); 
  mpz_init(X1); 
  mpz_init(Y1); 
  mpz_init(D1X); 
  mpz_init(D2X); 
  mpz_init(D1Y); 
  mpz_init(D2Y); 
  mpz_init (X);
  mpz_init (Y);
  mpz_init (M);
  mpz_init (bnd1);

  deriv_pol1 (P1X, P1, d, alpha); 
  deriv_pol1 (P2X, P2, d, alpha); 
  deriv_pol2 (P1Y, P1, d, alpha); 
  deriv_pol2 (P2Y, P2, d, alpha); 

  r = roots_mod_p (P1, P2, &p, &nb, d, alpha); 
  if (p==0) success=-1; 
  else success = 1;

  Nb[0] = nb;

  if ((success==1) && (nb >0)) {
    Sol[0] = (mpz_t *) malloc(nb*sizeof(mpz_t));
    Sol[1] = (mpz_t *) malloc(nb*sizeof(mpz_t));
    
    for (j=0; j< nb; j++) {
      mpz_init (Sol[0][j]);
      mpz_init (Sol[1][j]);
    }
    
    mpz_mul_ui (bnd1, bnd, 2);
    /*We apply Hensel lifting for all the roots mod p */
    
    
    for (j = 0; j < nb; j++) {
#ifdef DEBDEB	
      printf("p = %d, starting from x = %d, y = %d\n", 
	     p, r[j][0], r[j][1]); 
#endif
      
      mpz_set_ui (X, r[j][0]); 
      mpz_set_ui (Y, r[j][1]); 
      mpz_set_ui (M, p);     

      while (mpz_cmp (M, bnd1) <=0) {
	mpz_mul(M, M, M); 
	eval_pol (X1, P1, X, Y, M, d, alpha); 
	eval_pol (Y1, P2, X, Y, M, d, alpha); 
	eval_pol (D1X, P1X, X, Y, M, d, alpha); 
	eval_pol (D2X, P2X, X, Y, M, d, alpha);
	eval_pol (D1Y, P1Y, X, Y, M, d, alpha); 
	eval_pol (D2Y, P2Y, X, Y, M, d, alpha); 
	
#ifdef DEBDEB
	gmp_printf ("(X1, Y1) = (%Zd, %Zd)\n", X1, Y1); 
	gmp_printf ("(D1X, D1Y) = (%Zd, %Zd)\n", D1X, D1Y); 
	gmp_printf ("(D2X, D2Y) = (%Zd, %Zd)\n", D2X, D2Y); 
#endif
	
	mpz_mul (tmp1, D1X, D2Y); 
	mpz_fdiv_r (tmp1, tmp1, M); 
	
	mpz_mul (tmp2, D1Y, D2X); 
	mpz_fdiv_r (tmp2, tmp2, M); 
	
	mpz_sub (tmp1, tmp1, tmp2); 
	mpz_fdiv_r (tmp1, tmp1, M); 
	mpz_invert (tmp1, tmp1, M); 
	
#ifdef DEBDEB
	gmp_printf ("det = %Zd\n", tmp1); 
#endif
	
	mpz_mul (tmp2, D2Y, X1); 
	mpz_mul (tmp3, D1Y, Y1); 
	mpz_sub (tmp2, tmp2, tmp3); 
	mpz_mul (tmp2, tmp2, tmp1); 
	mpz_fdiv_r (tmp2, tmp2, M); 
	mpz_sub (X, X, tmp2); 
	mpz_fdiv_r (X, X, M); 
	
#ifdef DEBDEB
	gmp_printf ("(D2Y, D1Y) = (%Zd, %Zd)\n", D2Y, D1Y); 
	gmp_printf ("(X1, Y1) = %Zd %Zd\n", X1, Y1); 
#endif
	
	mpz_mul (tmp2, D1X, Y1); 
	mpz_mul (tmp3, D2X, X1); 
	mpz_sub (tmp2, tmp2, tmp3); 
	mpz_mul (tmp2, tmp2, tmp1); 
	mpz_fdiv_r (tmp2, tmp2, M); 
	mpz_sub (Y, Y, tmp2); 
	mpz_fdiv_r (Y, Y, M); 
      }

      /* Set tmp1 to ceil(M/2), and compute X cmod M and Y cmod M */
      
#ifdef DEBDEB
      gmp_printf("-> X = %Zd Y = %Zd M = %Zd bnd = %Zd\n", X, Y, M, bnd); 
#endif
      
      mpz_add_ui (tmp1, M, 1);
      mpz_div_ui (tmp1, M, 2);
      if (mpz_cmp (X, tmp1) >= 0) mpz_sub (X, X, M);
      if (mpz_cmp (Y, tmp1) >= 0) mpz_sub (Y, Y, M);
      
      mpz_set (Sol[0][j], X);
      mpz_set (Sol[1][j], Y);
#ifdef DEBDEB
      gmp_printf("-> X = %Zd Y = %Zd\n", X, Y); 
#endif
    }

    for (j = 0; j < p*p; j++) free(r[j]); 
    free(r);
  }


  for (j=0; j<=ncoefs-1; j++)
    {
      mpz_clear (P1X[j]);
      mpz_clear (P2X[j]);
      mpz_clear (P1Y[j]);
      mpz_clear (P2Y[j]);
    }
  
  free (P1X);
  free (P2X);
  free (P1Y);
  free (P2Y);

  mpz_clear (tmp1); 
  mpz_clear(tmp2); 
  mpz_clear(tmp3); 
  mpz_clear(X1); 
  mpz_clear(Y1); 
  mpz_clear(D1X); 
  mpz_clear(D2X); 
  mpz_clear(D1Y); 
  mpz_clear(D2Y); 
  mpz_clear (X);
  mpz_clear (Y);
  mpz_clear (M);
  mpz_clear (bnd1);

  return success;
}





/* Compute the dim-dimensional L1 norm of V and put it in tmp. */
void
Norm1 (mpz_t tmp, mpz_t *V, int dim)
{
  int i;

  mpz_set_ui (tmp, 0);
  for (i=1; i<=dim; i++) {
    if ((mpz_cmp_ui (V[i], 0)) >= 0) mpz_add (tmp, tmp, V[i]);
    else mpz_sub (tmp, tmp, V[i]);
  }
}




/**************************************************************************/
/**************************************************************************/
/* SLZ: from a polynomial approximation to a set of possible solutions,
   via lattice reduction and Hensel lifting */
/**************************************************************************/
/**************************************************************************/

/* Look for the solutions of the equation
   |P(x_0/T) cmod 1| < 1/C, 
   where  |x0| \leq T and C = floor( 1/(1/M+eps)).*/

int
slz (mpz_t **Res, int *Nb, mpfr_t *P, mpz_t C, mpz_t T, 
	int d, int alpha, mpfr_t tmp, mat_ZZ *B, mat_ZZ *G, 
	mat_ZZ *U, mat_ZZ *W, int npols, int ncoefs, 
	mpz_t ztmp, mpz_t Calpha, mpz_t ztmp2, int *ok, mpz_t *PP, 
	mpz_t *Ptmp)
{
  int dd= (d+3)/2;

#ifndef PROVED
  U->NumRows = G->NumCols; /*makes the compiler shut up. */
#endif

  int i, j, k, l, dx, dy, res;
  mpz_t *Sol[2];

  Nb[0] = 0;

#ifdef DEB
  printf ("\n\n*****ENTERING slz\n");
  printf ("ncoefs=%d, npols=%d\n", ncoefs, npols);
  printf ("C:="); mpz_out_str (stdout, 10, C); printf (";\n");
#endif

  /* Set PP= \lceil C*P \rfloor */
  for (i=0; i<=d; i++) {
    mpfr_mul_z (tmp, P[i], C, GMP_RNDN);
    mpfr_get_z (PP[i], tmp, GMP_RNDN);
  }

  /* |P(x_0/T) cmod 1| < 1/C  ==> PP(x_0/T) + y0 cmod C =0   
   where  |x0| \leq T and |y_0| \leq (d+3)/2 = dd.*/
  
#ifdef DEB
  printf ("\n**Starting filling the matrix**\n");
#endif

  /* FILL THE MATRIX B, ROW AFTER ROW */

  k=1;
  
  mpz_set (W->coeff[1][0], Calpha);
  for (i=1; i<ncoefs; i++)
    mpz_set_ui (W->coeff[1][i], 0);
  
  for (j=0; j<=alpha; j++) {
    /* W1 is (PP(x)+[(d+3)/2]y)^jC^(alpha-j).
       When j increases, W1 := W1 * (PP(x)+[(d+3)/2] y) / C 
    */

    mpz_set_ui (ztmp, 1);     /* Here, ztmp stores T^i */

    for (i=0; i<=alpha-j; i++) {
      /* Fill row nr k of B, ie, the one 
	 corresponding to x^i (PP(x)+[(d+3)/2]y)^jC^(alpha-j).
	 We have to shift W1 by X^i.      
      */

#ifdef DEB2
      printf("Matrix when starting j=%d, i=%d: \n", j, i);
      Print_mat (B->coeff, B->NumCols, B->NumRows); 
#endif

      for (dy=0; dy<=alpha; dy++)
	for (dx=0; dx <=d*(alpha-dy); dx++) {
	  if (dx<i) {
#ifdef VERIF
	    if (dx+(dy*(d*(2*alpha-dy+1)+2))/2+1 > ncoefs) abort();
	    if (dx+(dy*(d*(2*alpha-dy+1)+2))/2+1 <=0 ) abort();
#endif
	    mpz_set_ui(B->coeff[k][dx+(dy*(d*(2*alpha-dy+1)+2))/2+1],0);
	  }
	  else {
#ifdef VERIF
	    if (dx+(dy*(d*(2*alpha-dy+1)+2))/2+1 > ncoefs) abort();
	    if (dx+(dy*(d*(2*alpha-dy+1)+2))/2+1 <=0 ) abort();
	    if (dx-i+(dy*(d*(2*alpha-dy+1)+2))/2 <0) abort();
	    if (dx-i+(dy*(d*(2*alpha-dy+1)+2))/2 >= ncoefs) abort();
	    printf ("coeff = %d\n", dx+(dy*(d*(2*alpha-dy+1)+2))/2+1);
#endif
	    mpz_mul (B->coeff[k][dx+(dy*(d*(2*alpha-dy+1)+2))/2+1], 
		     ztmp, W->coeff[1][dx-i+(dy*(d*(2*alpha-dy+1)+2))/2]);
	  }
	}
      k++;
      mpz_mul (ztmp, ztmp, T);
    }

    if (j < alpha) { /* otherwise we are at the end of the loop */
      for (i=0; i<ncoefs; i++) mpz_set_ui (Ptmp[i], 0);
      
      for (l=0; l<=d; l++) {
	/*W2:=PP[l]*x^l*W1;  Ptmp += W2 */      
	
	for (dy=0; dy<=alpha; dy++)
	  for (dx=0; dx <=d*(alpha-dy); dx++){
	    if (dx<l) mpz_set_ui(W->coeff[2][dx+(dy*(d*(2*alpha-dy+1)+2))/2],0);
	    else {
	      mpz_set (W->coeff[2][dx+(dy*(d*(2*alpha-dy+1)+2))/2], 
		       W->coeff[1][dx-l+(dy*(d*(2*alpha-dy+1)+2))/2]);
	    }
	  }
	
	for (i=0; i<ncoefs; i++){
	  mpz_mul (W->coeff[2][i], W->coeff[2][i], PP[l]);
	  mpz_add (Ptmp[i], Ptmp[i], W->coeff[2][i]);
	}
      }
      
#ifdef DEB2
      for (i = 0; i<ncoefs; i++) {
	mpz_out_str (stdout, 10, Ptmp[i]);
	printf(" ");
      }
      printf("\n");
#endif
      
      /*W2:=[(d+3)/2]*y*W1;  Ptmp += W2 */
      
      for (dy=0; dy<=alpha; dy++)
	for (dx=0; dx <=d*(alpha-dy); dx++){
	  if (dy==0) mpz_set_ui (W->coeff[2][dx], 0);
	  else {
	    mpz_set (W->coeff[2][dx+(dy*(d*(2*alpha-dy+1)+2))/2], 
		     W->coeff[1][dx+((dy-1)*(d*(2*alpha-dy+2)+2))/2]);
	  }
	}
      
      for (i=0; i<ncoefs; i++)
	mpz_addmul_ui (Ptmp[i], W->coeff[2][i], dd);
      
#ifdef DEB2
      for (i = 0; i<ncoefs; i++) {
	mpz_out_str (stdout, 10, Ptmp[i]);
	printf(" ");
      }
      printf("\n");
#endif
      
      for (i=0; i<ncoefs; i++){
	mpz_set (W->coeff[1][i], Ptmp[i]);
	mpz_divexact (W->coeff[1][i], W->coeff[1][i], C);
      }
    }
  }



#ifdef DEB2
  printf("Matrix before LLL: \n");
  Print_mat (B->coeff, B->NumCols, B->NumRows); 
#endif

  /*************** LLL-reduce the matrix B ****************/

#ifdef PROVED
  Gram (G, B);
  proved_LLL (G, B, U);
#else
  fast_LLL (B);
#endif


#ifdef DEB  
  printf("\nMatrix after LLL: \n");
  Print_mat (B->coeff, B->NumCols, B->NumRows); 
#endif

  /* Take the first two vectors, look if their L1 norm are small (< C^alpha),
     and transform them into two polynomials 
     If it fails, take the first and the third, etc.

     ok[i] is -1 if the length of B->coeff[i] has not been computed yet,
     0 if it has been computed but it is too long or zero,
     1 if it has been computed and W[i] has also been computed

     We are looking for the common roots of W1 and W2 over Z 
     that are of the shape (tsol/T, ysol/[(d+3)/2]T^d), where tsol and ysol
     are integers satisfying |tsol| <= T, |ysol| <= [(d+3)/2] T^d.
     
     For Hensel's lifting, we require integer roots, so we change
     W1, W2 to T^(d*alpha) Wi(x/T,y/[(d+3)/2]T^d), 
     so that we will find tsol and ysol, of which only tsol is of interest.
     
     We did not perform this change of variables before lattice reduction
     because this would have been more expensive.
  */


  res = -1;
  for (i=1; i<=npols; i++) ok[i]=-1;
  
  for (i = 1; (res==-1) && (i<npols); i++) {
    if (ok[i] == -1) {
      Norm1 (ztmp, B->coeff[i], ncoefs);
      if (mpz_cmp (ztmp, Calpha)>=0) ok[i] = 0;
      else if (mpz_cmp_ui (ztmp, 0)==0) ok[i] = 0;
      else ok[i]=1;
#ifdef DEB    
      printf ("i=%d, ok[%d]=%d\n", i, i, ok[i]);
#endif  
      if (ok[i] == 1) {
	for (k=0; k<ncoefs; k++) 
	  mpz_set (W->coeff[i][k], B->coeff[i][k+1]);
	
#ifdef DEB2
	printf("This is W[%d]: \n", i);
	for (k = 0; k<ncoefs; k++) {
	  mpz_out_str (stdout, 10, W->coeff[i][k]);
	  printf(" ");
      }
      printf("\n");
#endif
      
	mpz_pow_ui (ztmp, T, d*alpha);
	for (k=0; k<ncoefs; k++) 
	  mpz_mul (W->coeff[i][k], W->coeff[i][k], ztmp);
	
	for (dy=0; dy<=alpha; dy++)
	  for (dx=0; dx <=d*(alpha-dy); dx++){
	    /* divide x^dx y^dy by T^(dx+d*dy)*[(d+3)/2]^dy */
	    mpz_pow_ui (ztmp, T, dx+d*dy);
	    mpz_set_ui (ztmp2, dd);
	    mpz_pow_ui (ztmp2, ztmp2, dy);
	    mpz_mul (ztmp, ztmp, ztmp2);
	    mpz_divexact (W->coeff[i][dx+(dy * (d*(2*alpha-dy+1)+2))/2], 
			  W->coeff[i][dx+(dy * (d*(2*alpha-dy+1)+2))/2], 
			  ztmp);
	  }
	
#ifdef DEB2
	printf("This is W[%d] \n", i);
	for (k = 0; k<ncoefs; k++) {
	  mpz_out_str (stdout, 10, W->coeff[i][k]);
	  printf(" ");
	}
	printf("\n");
#endif
	
      }
    }
    if (ok[i]==0) continue;
    
    for (j = i+1; (res==-1) && (j<=npols); j++) {
#ifdef DEB
      printf ("j=%d\n", j);
#endif      
      if (ok[j] == -1) {
	Norm1 (ztmp, B->coeff[j], ncoefs);
	if (mpz_cmp (ztmp, Calpha)>=0) ok[j]=0;
	else if (mpz_cmp_ui (ztmp, 0) ==0) ok[j] = 0;
	else ok[j] = 1;

#ifdef DEB    
	printf ("j=%d, ok[%d]=%d\n", j, j, ok[j]);
#endif  
	
	if (ok[j] == 1) {
	  for (k=0; k<ncoefs; k++) 
	    mpz_set (W->coeff[j][k], B->coeff[j][k+1]);
	  
	  mpz_pow_ui (ztmp, T, d*alpha);
	  for (k=0; k<ncoefs; k++) 
	    mpz_mul (W->coeff[j][k], W->coeff[j][k], ztmp);
	  
	  for (dy=0; dy<=alpha; dy++)
	    for (dx=0; dx <=d*(alpha-dy); dx++){
	      /* divide x^dx y^dy by T^(dx+d*dy)*[(d+3)/2]^dy */
	      mpz_pow_ui (ztmp, T, dx+d*dy);
	      mpz_set_ui (ztmp2, dd);
	      mpz_pow_ui (ztmp2, ztmp2, dy);
	      mpz_mul (ztmp, ztmp, ztmp2);
	      mpz_divexact (W->coeff[j][dx+(dy * (d*(2*alpha-dy+1)+2))/2], 
			    W->coeff[j][dx+(dy * (d*(2*alpha-dy+1)+2))/2], 
			    ztmp);
	    }

#ifdef DEB2
	printf("This is W[%d] \n", j);
	for (k = 0; k<ncoefs; k++) {
	  mpz_out_str (stdout, 10, W->coeff[j][k]);
	  printf(" ");
	}
	printf("\n");
#endif
	
	}
      }
      if (ok[j]==0) continue;
      
#ifdef DEB
      printf ("**Passed the norm conditions for i=%d and j=%d\n", i, j);
#endif 
      
      mpz_pow_ui (ztmp, T, d);
      mpz_mul_ui (ztmp, ztmp, dd);
      mpz_add_ui (ztmp, ztmp, 1);
      res = hensel (Sol, Nb, W->coeff[i], W->coeff[j], ztmp, d, alpha); 
      
#ifdef DEB2
      printf ("Nb[0] is %d\n", Nb[0]);
      if (Nb[0]>0) gmp_printf ("Sol[0][0] is %Zd\n", Sol[0][0]);
#endif
      
    }
  }

  if ((res==1)&&(Nb[0] > 0)) {
    Res[0] = (mpz_t *) malloc(Nb[0]*sizeof(mpz_t)); 
    
    for (j=0; j< Nb[0]; j++) {
      mpz_init (Res[0][j]);
      mpz_set (Res[0][j], Sol[0][j]);
      mpz_clear (Sol[0][j]);
      mpz_clear (Sol[1][j]);
    }      
    free (Sol[0]);
    free (Sol[1]);
  }


#ifdef DEB2
      printf ("Nb[0] is %d\n", Nb[0]);
      if (Nb[0] > 0) gmp_printf ("Res[0][0] is %Zd\n", Res[0][0]);
#endif


#ifdef DEB
  printf ("*****EXITING slz*****\n\n\n");
#endif  

  return res;
}




/********************************************************************/
/* checks that the retuned possible solutions in  Sol are indeed 
   solutions, in which case they are printed */
/********************************************************************/

/* ie, tests for all x0 \in Res if:
   | NN. f( (x0+tm)/N ) cmod 1| \leq 1/M */
int
Check (mpz_t *Res, int Nb, mpz_t tm, mpz_t N, mpfr_t NN, 
       mpfr_t oneoverM, mpz_t T, int e_in)
{
  int i, k = 0;
  mpz_t x1;
  mpfr_t y1, y2;

  mpz_init (x1);
  mpfr_init (y1);
  mpfr_init (y2);

  for (i=0; i<Nb; i++)  {
    /*check if |Res[i]|<=T  */

    mpz_abs (x1, Res[i]);
    if ( mpz_cmp (x1, T) <=0 ){

      /*compute cfrac(NN*exp[ (Res[i]+tm)/N] )   */
      mpz_add (Res[i], Res[i], tm);
      mpfr_set_z (y1, Res[i], GMP_RNDN);
      mpfr_div_z (y1, y1, N, GMP_RNDN);

#if FUNCTION==EXPO
      mpfr_exp (y1, y1, GMP_RNDN);    
#elif FUNCTION==TWOTOX
      mpfr_exp2 (y1, y1, GMP_RNDN);
#elif FUNCTION==LOG2 
      mpfr_log2 (y1, y1, GMP_RNDN);
#elif FUNCTION==LOG10
      mpfr_log10 (y1, y1, GMP_RNDN);
#elif FUNCTION==LN
      mpfr_log (y1, y1, GMP_RNDN);
#elif FUNCTION==SIN
      mpfr_sin (y1, y1, GMP_RNDN);
#elif FUNCTION==COS
      mpfr_cos (y1, y1, GMP_RNDN);
#elif FUNCTION==ONEOVERX
      mpfr_ui_div (y1, 1, y1, GMP_RNDN);
#elif FUNCTION==HALL
      mpfr_pow_ui (y1, y1, 3, GMP_RNDN);
      mpfr_sqrt (y1, y1, GMP_RNDN);
#elif FUNCTION==SQROOT
      mpfr_sqrt (y1, y1, GMP_RNDN);
#elif FUNCTION==SQRTINVERT
      mpfr_sqrt (y1, y1, GMP_RNDN);
      mpfr_ui_div (y1, 1, y1, GMP_RNDN);
#endif


      mpfr_mul (y1, y1, NN, GMP_RNDN);

      if (near==1) mpfr_sub (y1, y1, target, GMP_RNDN);

      mpfr_round (y2, y1);
      mpfr_sub (y1, y1, y2, GMP_RNDN);
      mpfr_abs (y1, y1, GMP_RNDN);
      
      /*check if y1 < 1/M */
      
#ifdef DEB2
      printf ("oneover M is ");
      mpfr_out_str (stdout, 10, 10, oneoverM, GMP_RNDN);
      printf ("\n");
      printf ("y1 is ");
      mpfr_out_str (stdout, 10, 10, y1, GMP_RNDN);
      printf ("\n");
#endif
      
      if ( (mpfr_cmp (y1, oneoverM) <= 0) )  {
	/* PRINT THE FOUND SOLUTION */
	k ++;
	printf ("*** x=");
	mpz_out_str (stdout, 10, Res[i]);
	printf("E%d, distance is: ", e_in);
	mpfr_out_str (stdout, 10, 10, y1, GMP_RNDN);
	printf ("\n");
	fflush (stdout);
      }
    }
  }

  mpz_clear (x1);
  mpfr_clear (y1);
  mpfr_clear (y2);

  return k;
}




/**************************************************************************/

/*bacsel nearest 100 8 8 8.0 2.0 50000000 100000000 1 1 [e_in] [e_out]*/

int
main (int argc, char *argv[])
{
  mp_prec_t prec;
  int i, j, k, d, alpha, Nb, success, number=0;
  double n, nn, m, t;
  mpz_t initial_T, t0, t1, N, T, tm, *Res=NULL, C, initial_C, 
    ztmp, ztmp2, Calpha;
  mpfr_t tmp, eps, *P, M, oneoverM, initial_eps, Mprime, NN;
  long long int failures = 0, good=0;     /* number of failures */
  double total_T = 0.0;                   /* total width */
  int e_in, e_out, npols, ncoefs;
#ifdef VBSE
  double testtmp, t0init, total_length;
  double test_i=10.0;
#endif
  mat_ZZ B, G, W, U;
  int *ok;
  mpz_t *PP;
  mpz_t *Ptmp;





  fprintf (stderr, "%s ", argv[0]);
  {
    int i;
    for (i = 1; i < argc; i++)
      fprintf (stderr, "%s ", argv[i]);
    fprintf (stderr, "\n");
  }

  fprintf (stderr, "Welcome to BaCSeL version %s\n", VERSION);

  

  /* ARGUMENTS CONTROL */

  if (argc < 11 || argc > 13) {
    fprintf (stderr, 
	     "Usage: ./bacsel rnd_mode prec n nn m t t0 t1 d alpha [e_in] [e_out]\n");
    exit (1);
  }


  /* prec = working precision, 
     n = input precision,
     nn= output precision,
     m = quality of the bad cases,
     T=basis^(t-n) = size of the sub-intervals,
     t0= starting fp number is t0/basis^n,
     t1= arriving fp number is t1/basis^n.
     e_in [optional] input exponent  (default = 0)
     e_out [optional] output exponent (default = 0)

     ---> looking for the solutions of the equation
        | basis^nn . basis^(-e_out).f(t.basis^(e_in)/basis^n) cmod 1| \leq basis^-m, 
        | t in [|t0, t1|]

     nn, n, m, t are doubles.
     d, alpha are ints: these are parameters of the method.
     Here, f(x)=exp(x).
  */

  /* INITIALIZING n, nn, m, t, initial_T, t0, t1, tm, d, alpha */

  if ((strcmp(argv[1],"directed"))==0) near=0;
  else if ((strcmp(argv[1],"nearest"))==0) near=1;
  else if ((strcmp(argv[1],"all"))==0) near=2;
  else {
    fprintf (stderr, "The 1st argument must be chosen among directed, nearest and all.\n");
    exit (0);
  }

  prec = atoi (argv[2]);
  mpfr_set_default_prec (prec);

  n  = atof (argv[3]);
  nn = atof (argv[4]);
  m  = atof (argv[5]);
  t  = atof (argv[6]);

  mpz_init (t0);
  mpz_init (t1);
  mpz_set_str (t0, argv[7], 10);
  mpz_set_str (t1, argv[8], 10);

#ifdef VBSE
  t0init = mpz_get_d(t0);
  total_length = mpz_get_d(t1) - t0init;
#endif

  d = atoi (argv[9]);
  alpha = atoi (argv[10]);
  e_in = (argc>=12)? atoi (argv[11]): 0;
  e_out = (argc==13)? atoi (argv[12]): 0;

  mpfr_init (target);
  mpfr_set_d (target, 0.5, GMP_RNDN);

  mpz_init (initial_T);
  mpz_init (T);
  mpz_init (tm);
  mpz_init (ztmp);
  mpz_init (ztmp2);
  mpz_init (Calpha);



  mpz_init (N);
  mpfr_init (NN);

  mpfr_init (tmp);
  mpfr_init (M);
  mpfr_init (oneoverM);
  mpfr_init (Mprime);
  mpz_init (C);
  mpz_init (initial_C);
  mpfr_init (eps);
  mpfr_init (initial_eps);



#ifdef VBSE
  fprintf (stderr, 
	   "You are looking for %e-bad cases for the \n",
	   m);
#if FUNCTION==EXPO
  fprintf (stderr,
	   "rounding of exp(x), over [%e,%e] with:\n",
	   (mpz_get_d (t0))/ pow((double) BASIS, n-e_in), 
	   (mpz_get_d (t1))/ pow((double) BASIS, n-e_in));
#elif FUNCTION==TWOTOX
  fprintf (stderr,
	   "rounding of 2^x, over [%e,%e] with:\n",
	   (mpz_get_d (t0))/ pow((double) BASIS, n-e_in), 
	   (mpz_get_d (t1))/ pow((double) BASIS, n-e_in));
#elif FUNCTION==LOG2
  fprintf (stderr,
	   "rounding of log[2](x), over [%e,%e] with:\n",
	   (mpz_get_d (t0))/ pow((double) BASIS, n-e_in), 
	   (mpz_get_d (t1))/ pow((double) BASIS, n-e_in));
#elif FUNCTION==LOG10
  fprintf (stderr,
	   "rounding of log[10](x), over [%e,%e] with:\n",
	   (mpz_get_d (t0))/ pow((double) BASIS, n-e_in), 
	   (mpz_get_d (t1))/ pow((double) BASIS, n-e_in));
#elif FUNCTION==LN
  fprintf (stderr,
	   "rounding of ln(x), over [%e,%e] with:\n",
	   (mpz_get_d (t0))/ pow((double) BASIS, n-e_in), 
	   (mpz_get_d (t1))/ pow((double) BASIS, n-e_in));
#elif FUNCTION==SIN
  fprintf (stderr,
	   "rounding of sin(x), over [%e,%e] with:\n",
	   (mpz_get_d (t0))/ pow((double) BASIS, n-e_in), 
	   (mpz_get_d (t1))/ pow((double) BASIS, n-e_in));
#elif FUNCTION==COS
  fprintf (stderr,
	   "rounding of cos(x), over [%e,%e] with:\n",
	   (mpz_get_d (t0))/ pow((double) BASIS, n-e_in), 
	   (mpz_get_d (t1))/ pow((double) BASIS, n-e_in));
#elif FUNCTION==ONEOVERX
  fprintf (stderr,
	   "rounding of 1/x, over [%e,%e] with:\n",
	   (mpz_get_d (t0))/ pow((double) BASIS, n-e_in), 
	   (mpz_get_d (t1))/ pow((double) BASIS, n-e_in));
#elif FUNCTION==HALL
  fprintf (stderr,
	   "rounding of x^(3/2), over [%e,%e] with:\n",
	   (mpz_get_d (t0))/ pow((double) BASIS, n-e_in), 
	   (mpz_get_d (t1))/ pow((double) BASIS, n-e_in));
#elif FUNCTION==SQROOT
  fprintf (stderr,
	   "rounding of x^(1/2), over [%e,%e] with:\n",
	   (mpz_get_d (t0))/ pow((double) BASIS, n-e_in), 
	   (mpz_get_d (t1))/ pow((double) BASIS, n-e_in));
#elif FUNCTION==SQRTINVERT
  fprintf (stderr,
	   "rounding of x^(-1/2), over [%e,%e] with:\n",
	   (mpz_get_d (t0))/ pow((double) BASIS, n-e_in), 
	   (mpz_get_d (t1))/ pow((double) BASIS, n-e_in));
#endif
  if (near==0) fprintf (stderr, " -rounding mode: directed\n");
  if (near==1) fprintf (stderr, " -rounding mode: nearest\n");  
  if (near==2) fprintf (stderr, " -rounding mode: all\n");
  fprintf (stderr,
	   " -input precision: %e base-%d digits\n",
	   n, BASIS);
  fprintf (stderr,
	   " -output precision: %e base-%d digits\n",
	   nn, BASIS);
  fprintf (stderr,
	   " -working precision: %d bits\n",
	   atoi (argv[2]));
  fprintf (stderr,
	   " -degree of approximating polynomials: %d\n",
	   d);
  fprintf (stderr,
	   " -parameter for Coppersmith's technique: %d\n",
	   alpha);
  fprintf (stderr,
	   " -size of subintervals: %e\n", 
	   exp(log(BASIS)*(t-n)));
  fprintf (stderr,
	   " -e_in=%d, e_out=%d \n \n", 
	   e_in, e_out);
  fprintf (stderr, "Dimension of the lattices is %dx%d\n", 
	   ((alpha+1)*(alpha+2))/2, ((alpha+1)*(d*alpha+2))/2);
#endif  


  /* N = basis^n, NN=basis^nn, M=basis^m, T=initial_T=basis^t */
  mpfr_set_d (tmp, n, GMP_RNDN);
  mpfr_ui_pow (tmp, BASIS, tmp, GMP_RNDN);
  mpfr_get_z (N, tmp, GMP_RNDN);

  if (mpz_sgn (t0) != mpz_sgn (t1)){
    fprintf (stderr, "t0 and t1 must have the same sign.\n");
    exit (0);
  }

  if (mpz_cmp(t0, t1)>=0){
    fprintf (stderr, "We must have t0<t1.\n");
    exit (0);
  }

  if (mpz_sgn(t0)>0){
    if (mpz_cmp (t1, N)>0){
      fprintf (stderr, "t1 is too big, the computation will stop at t1=");
      mpz_out_str (stderr, 10, N);
      fprintf (stderr, "\n");
      mpz_set (t1, N);
    }
    
    mpz_div_ui (C, N, BASIS);
    if (mpz_cmp (C, t0)>0){
      fprintf (stderr, "t0 is too small, the computation will start at t0=");
      mpz_out_str (stderr, 10, N);
      fprintf (stderr, "\n");
      mpz_set (t0, N);
    }
  }
  else{
    mpz_abs (T, t0);
    if (mpz_cmp (T, N)>0){
      fprintf (stderr, "t0 is too small, the computation will start at t0=-");
      mpz_out_str (stderr, 10, N);
      fprintf (stderr, "\n");
      mpz_set (t0, N);
      mpz_neg (t0, t0);
    }
    
    mpz_div_ui (C, N, BASIS);
    mpz_abs (T, t1);
    if (mpz_cmp (C, T)>0){
      fprintf (stderr, "t1 is too large, the computation will stop at t1=-");
      mpz_out_str (stderr, 10, C);
      fprintf (stderr, "\n");
      mpz_set (t1, C);
      mpz_neg (t1, t1);
    }
  }
  
  /* From now on, N is Nold * basis^(-e_in). 
     Then everything remains as without e_in.*/


  mpz_set_ui (T, 1);
  for (i=1; i<=abs(e_in); i++) mpz_mul_ui (T, T, BASIS);
  (e_in>0)? mpz_div (N, N, T) : mpz_mul (N, N, T);

  mpfr_set_d (tmp, nn, GMP_RNDN);
  mpfr_ui_pow (NN, BASIS, tmp, GMP_RNDN);
  mpfr_set_si (tmp, BASIS, GMP_RNDN);
  mpfr_pow_si (tmp, tmp, -e_out, GMP_RNDN);
  mpfr_mul (NN, NN, tmp, GMP_RNDN);
  if (near == 2) mpfr_mul_ui (NN, NN, 2, GMP_RNDN);

  mpfr_set_d (tmp, m, GMP_RNDN);
  mpfr_ui_pow (M, BASIS, tmp, GMP_RNDN);
  mpfr_ui_div (oneoverM, 1, M, GMP_RNDU);

  mpfr_set_d (tmp, t, GMP_RNDN);
  mpfr_ui_pow (tmp, BASIS, tmp, GMP_RNDN);
  mpfr_get_z (T, tmp, GMP_RNDN);
  mpz_set (initial_T, T);

#ifdef DEB2
  printf("N= ");
  mpz_out_str (stdout, 10, N);
  printf ("\n"); 
  printf("NN= ");
  mpfr_out_str (stdout, 10, 10, NN, GMP_RNDN);
  printf ("\n"); 
  printf("M= ");
  mpfr_out_str (stdout, 10, 10, M, GMP_RNDN);
  printf ("\n"); 
#endif

  /* Bounding eps once for all: 
     eps <= max_{t0/N..t1/N} [ |(d^(d+1) f)/(dx)^(d+1) | ] 
            * 1/(d+1)! * [normNN*T^(d+1)/normN^(d+1)] */

  mpfr_set_z (eps, T, GMP_RNDU);
  mpfr_pow_ui (eps, eps, d+1, GMP_RNDU);
  mpfr_set_z (tmp, N, GMP_RNDD);
  mpfr_pow_ui (tmp, tmp, d+1, GMP_RNDD);
  mpfr_div (tmp, eps, tmp, GMP_RNDU);
  mpfr_mul (eps, tmp, NN, GMP_RNDU);

  for (i = 2; i<=d+1; i++)
    mpfr_div_ui (eps, eps, i, GMP_RNDU);

  /* eps is now 1/(d+1)! * [normNN*T^(d+1)/normN^(d+1)] */
  /* At this point, t0 and t1 have the same sign */

#if FUNCTION==EXPO   
  /* exp(x) */
  mpfr_set_z (tmp, t1, GMP_RNDU);
  mpfr_div_z (tmp, tmp, N, GMP_RNDU);
  mpfr_exp (tmp, tmp, GMP_RNDU);
#elif FUNCTION==TWOTOX   
    /* 2^x * (ln(2)^(d+1)) */
  mpfr_set_z (tmp, t1, GMP_RNDU);
  mpfr_div_z (tmp, tmp, N, GMP_RNDU);
  mpfr_exp2 (tmp, tmp, GMP_RNDU);
  mpfr_set_ui (initial_eps, 2, GMP_RNDU);
  mpfr_log (initial_eps, initial_eps, GMP_RNDU);
  mpfr_pow_ui (initial_eps, initial_eps, d+1, GMP_RNDU);
  mpfr_mul (tmp, tmp, initial_eps, GMP_RNDU);
#elif FUNCTION==LOG2    
  /* d! / x^(d+1)  * 1/ln(2) */
  mpfr_set_z (tmp, t1, GMP_RNDU);
  mpfr_div_z (tmp, tmp, N, GMP_RNDD);
  mpfr_ui_div (tmp, 1, tmp, GMP_RNDU);
  for (i=1; i<=d; i++)
    mpfr_mul_ui (tmp, tmp, i, GMP_RNDU);
  mpfr_set_ui (initial_eps, 2, GMP_RNDD);
  mpfr_log (initial_eps, initial_eps, GMP_RNDD);
  mpfr_ui_div (initial_eps, 1, initial_eps, GMP_RNDU);
  mpfr_mul (tmp, tmp, initial_eps, GMP_RNDU);
#elif FUNCTION==LOG10
  /* d! / x^(d+1)  * 1/ln(10) */
  mpfr_set_z (tmp, t1, GMP_RNDU);
  mpfr_div_z (tmp, tmp, N, GMP_RNDD);
  mpfr_ui_div (tmp, 1, tmp, GMP_RNDU);
  for (i=1; i<=d; i++)
    mpfr_mul_ui (tmp, tmp, i, GMP_RNDU);
  mpfr_set_ui (initial_eps, 10, GMP_RNDD);
  mpfr_log (initial_eps, initial_eps, GMP_RNDD);
  mpfr_ui_div (initial_eps, 1, initial_eps, GMP_RNDU);
  mpfr_mul (tmp, tmp, initial_eps, GMP_RNDU);
#elif FUNCTION==LN
  /* d! / x^(d+1)  * 1/ln(10) */
  mpfr_set_z (tmp, t1, GMP_RNDU);
  mpfr_div_z (tmp, tmp, N, GMP_RNDD);
  mpfr_ui_div (tmp, 1, tmp, GMP_RNDU);
  for (i=1; i<=d; i++)
    mpfr_mul_ui (tmp, tmp, i, GMP_RNDU);
#elif FUNCTION==SIN
  /* 1 */
  mpfr_set_ui (tmp, 1, GMP_RNDU);
#elif FUNCTION==COS
  /* 1 */
  mpfr_set_ui (tmp, 1, GMP_RNDU);
#elif FUNCTION==ONEOVERX
  /* (d+1)! / x^(d+2) */
  if (mpz_sgn(t0)>0){
    mpfr_set_z (tmp, t0, GMP_RNDU);
    mpfr_div_z (tmp, tmp, N, GMP_RNDD);
  }
  else{
    mpfr_set_z (tmp, t1, GMP_RNDD);
    mpfr_abs (tmp, tmp, GMP_RNDU);
    mpfr_div_z (tmp, tmp, N, GMP_RNDD);
    }
  mpfr_ui_div (tmp, 1, tmp, GMP_RNDU);
  mpfr_pow_ui (tmp, tmp, d+2, GMP_RNDU);
  for (i=1; i<=d+1; i++)
    mpfr_mul_ui (tmp, tmp, i, GMP_RNDU);
#elif FUNCTION==HALL
  /* [2^(-d-1) * 3.1.1.3....(2(d-2)+1) / x^((2d-1)/2) */
  mpfr_set_z (tmp, t0, GMP_RNDU);
  mpfr_div_z (tmp, tmp, N, GMP_RNDD);
  mpfr_ui_div (tmp, 1, tmp, GMP_RNDU);
  mpfr_set_d (initial_eps, ((double) (2*d-1))/2.0, GMP_RNDU);
  mpfr_pow (tmp, tmp, initial_eps, GMP_RNDU);
  for (i=1; i<=d-2; i++)
    mpfr_mul_ui (tmp, tmp, 2*i+1, GMP_RNDU);
  mpfr_mul_ui (tmp, tmp, 3, GMP_RNDU);
  mpfr_div_2exp (tmp, tmp, d+1, GMP_RNDU);
#elif FUNCTION==SQROOT
  /* [2^(-d-1) * 1.3....(2(d-1)+1) / x^((2d+1)/2) */
  mpfr_set_z (tmp, N, GMP_RNDD);
  mpfr_div_z (tmp, tmp, t0, GMP_RNDU);
  mpfr_set_d (initial_eps, ((double) d)+0.5, GMP_RNDU);
  mpfr_pow (tmp, tmp, initial_eps, GMP_RNDU);
  for (i=1; i<=d-1; i++)
    mpfr_mul_ui (tmp, tmp, 2*i+1, GMP_RNDU);
  mpfr_div_2exp (tmp, tmp, d+1, GMP_RNDU);
#elif FUNCTION==SQRTINVERT
  /* [2^(-d-1) * 1.3....(2d+1) / x^((2d+3)/2) */
  mpfr_set_z (tmp, t0, GMP_RNDU);
  mpfr_div_z (tmp, tmp, N, GMP_RNDD);
  mpfr_ui_div (tmp, 1, tmp, GMP_RNDU);
  mpfr_set_d (initial_eps, ((double) (2*d+3))/2.0, GMP_RNDU);
  mpfr_pow (tmp, tmp, initial_eps, GMP_RNDU);
  for (i=1; i<=d; i++)
    mpfr_mul_ui (tmp, tmp, 2*i+1, GMP_RNDU);
  mpfr_div_2exp (tmp, tmp, d+1, GMP_RNDU);
#endif

  mpfr_mul (eps, eps, tmp, GMP_RNDU);
  mpfr_set (initial_eps, eps, GMP_RNDU);

#ifdef DEB2
  printf("eps= ");
  mpfr_out_str (stdout, 10, 10, eps, GMP_RNDN);
  printf ("\n"); 
#endif

  /* Mprime := floor((1/2)/[1/M+eps]) 
     We take 1/2 because there are numerical errors on the coefficients
     of the approximating polynomials. 
     C:= Mprime
  */

  mpfr_ui_div (Mprime, 1, M, GMP_RNDN);
  mpfr_add (Mprime, Mprime, eps, GMP_RNDN);
  mpfr_ui_div (Mprime, 1, Mprime, GMP_RNDN);
  mpfr_div_ui (Mprime, Mprime, 2, GMP_RNDN);

  mpfr_get_z (C, Mprime, GMP_RNDD);
  mpz_set (initial_C, C);
  mpz_pow_ui (Calpha, C, alpha);

  ncoefs = (alpha + 1)*(d*alpha + 2)/2;
  npols = (alpha + 1)*(alpha + 2)/2;

  ok = (int *) malloc ((npols+1) * sizeof (int));
  PP = init_vec (d+1);
  Ptmp = init_vec (ncoefs);

  P = (mpfr_t *) malloc ((d+1)*sizeof (mpfr_t));
  for (i = 0; i<=d; i++) mpfr_init (P[i]);

  init_matrix (&B, npols, ncoefs);
  init_matrix (&W, npols, ncoefs);
  init_matrix (&U, 0, 0);

#ifdef PROVED
  init_matrix (&G, npols, npols);
#else
  init_matrix (&G, 0, 0);
#endif


  /* MAIN LOOP */

  while (mpz_cmp (t0, t1) < 0) {
#ifdef DEB1
    printf ("x=");
    mpz_out_str (stdout, 10, t0);
    printf ("\n");
#endif
    
    mpz_add (tm, t0, T);     /* tm is the middle of current interval */



    /*Polynomial approximation step*/
#if FUNCTION==EXPO
    mpfr_set_z (tmp, tm, GMP_RNDN);
    mpfr_div_z (tmp, tmp, N, GMP_RNDN);
    mpfr_exp (tmp, tmp, GMP_RNDN);
    /* compute series expansion 
       P(x)=NN*[exp(tm/normN) +exp(tm.2^e_in/normN)*T/normN*x+...
       +exp(tm.2^e_in/normN)/d! *T^d/normN^d*x^d]

       if |NN.f(x_0/normN) cmod 1| < 1/M, 
       then |P((x_0-tm).2^e_in/T) cmod 1| < 1/C,
       where  |x0-tm|.2^e_in \leq T and C = floor (1/ (1/M+eps)).
    */
    mpfr_mul (P[0], tmp, NN, GMP_RNDN);
    for (i=1; i<=d; i++) {
      mpfr_mul_z (P[i], P[i-1], T, GMP_RNDN);
      mpfr_div_z (P[i], P[i], N, GMP_RNDN);
      mpfr_div_ui (P[i], P[i], i, GMP_RNDN);
    }
#elif FUNCTION==TWOTOX
    mpfr_set_z (tmp, tm, GMP_RNDN);
    mpfr_div_z (tmp, tmp, N, GMP_RNDN);
    mpfr_exp2 (P[0], tmp, GMP_RNDN);
    mpfr_mul (P[0], P[0], NN, GMP_RNDN);
    mpfr_set_ui (tmp, 2, GMP_RNDN);
    mpfr_log (tmp, tmp, GMP_RNDN);
    for (i=1; i<=d; i++) {
      mpfr_mul_z (P[i], P[i-1], T, GMP_RNDN);
      mpfr_div_z (P[i], P[i], N, GMP_RNDN);
      mpfr_div_ui (P[i], P[i], i, GMP_RNDN);
      mpfr_mul (P[i], P[i], tmp, GMP_RNDN);
    }
#elif FUNCTION==LOG2
    mpfr_set_z (tmp, tm, GMP_RNDN);
    mpfr_div_z (tmp, tmp, N, GMP_RNDN);
    mpfr_log2 (P[0], tmp, GMP_RNDN);
    mpfr_mul (P[0], P[0], NN, GMP_RNDN);
    mpfr_ui_div (tmp, 1, tmp, GMP_RNDN);
    mpfr_set_ui (P[1], 2, GMP_RNDN);
    mpfr_log (P[1], P[1], GMP_RNDN);
    mpfr_ui_div (P[1], 1, P[1], GMP_RNDN);
    mpfr_mul (P[1], P[1], tmp, GMP_RNDN);
    mpfr_mul_z (P[1], P[1], T, GMP_RNDN);
    mpfr_div_z (P[1], P[1], N, GMP_RNDN);
    mpfr_mul (P[1], P[1], NN, GMP_RNDN);
    for (i=2; i<=d; i++) {
      mpfr_mul_si (P[i], P[i-1], -1, GMP_RNDN);
      mpfr_mul_z (P[i], P[i], T, GMP_RNDN);
      mpfr_div_z (P[i], P[i], N, GMP_RNDN);
      mpfr_mul (P[i], P[i], tmp, GMP_RNDN);
      mpfr_mul_ui (P[i], P[i], i-1, GMP_RNDN);
      mpfr_div_ui (P[i], P[i], i, GMP_RNDN);
    }
#elif FUNCTION==LOG10
  mpfr_set_z (tmp, tm, GMP_RNDN);
  mpfr_div_z (tmp, tmp, N, GMP_RNDN);
  mpfr_log10 (P[0], tmp, GMP_RNDN);
  mpfr_mul (P[0], P[0], NN, GMP_RNDN);
  mpfr_ui_div (tmp, 1, tmp, GMP_RNDN);
  mpfr_set_ui (P[1], 10, GMP_RNDN);
  mpfr_log (P[1], P[1], GMP_RNDN);
  mpfr_ui_div (P[1], 1, P[1], GMP_RNDN);
  mpfr_mul (P[1], P[1], tmp, GMP_RNDN);
  mpfr_mul_z (P[1], P[1], T, GMP_RNDN);
  mpfr_div_z (P[1], P[1], N, GMP_RNDN);
  mpfr_mul (P[1], P[1], NN, GMP_RNDN);
  for (i=2; i<=d; i++) {
    mpfr_mul_si (P[i], P[i-1], -1, GMP_RNDN);
    mpfr_mul_z (P[i], P[i], T, GMP_RNDN);
    mpfr_div_z (P[i], P[i], N, GMP_RNDN);
    mpfr_mul (P[i], P[i], tmp, GMP_RNDN);
    mpfr_mul_ui (P[i], P[i], i-1, GMP_RNDN);
    mpfr_div_ui (P[i], P[i], i, GMP_RNDN);
  }
#elif FUNCTION==LN
  mpfr_set_z (tmp, tm, GMP_RNDN);
  mpfr_div_z (tmp, tmp, N, GMP_RNDN);
  mpfr_log (P[0], tmp, GMP_RNDN);
  mpfr_mul (P[0], P[0], NN, GMP_RNDN);
  mpfr_ui_div (tmp, 1, tmp, GMP_RNDN);
  mpfr_set (P[1], tmp, GMP_RNDN);
  mpfr_mul_z (P[1], P[1], T, GMP_RNDN);
  mpfr_div_z (P[1], P[1], N, GMP_RNDN);
  mpfr_mul (P[1], P[1], NN, GMP_RNDN);
  for (i=2; i<=d; i++) {
    mpfr_mul_si (P[i], P[i-1], -1, GMP_RNDN);
    mpfr_mul_z (P[i], P[i], T, GMP_RNDN);
    mpfr_div_z (P[i], P[i], N, GMP_RNDN);
    mpfr_mul (P[i], P[i], tmp, GMP_RNDN);
    mpfr_mul_ui (P[i], P[i], i-1, GMP_RNDN);
    mpfr_div_ui (P[i], P[i], i, GMP_RNDN);
  }
#elif FUNCTION==SIN
  mpfr_set_z (tmp, tm, GMP_RNDN);
  mpfr_div_z (tmp, tmp, N, GMP_RNDN);
  mpfr_sin (P[0], tmp, GMP_RNDN);
  mpfr_mul (P[0], P[0], NN, GMP_RNDN);
  for (i=1; i<=d; i++) {
    if (i%2==0){
      mpfr_sin (P[i], tmp, GMP_RNDN);
      for (j=1; j<=i; j++)
	mpfr_div_ui (P[i], P[i], j, GMP_RNDN);
      if (i%4==2)
	mpfr_mul_si (P[i], P[i], -1, GMP_RNDN);
    }
    else{
      mpfr_cos (P[i], tmp, GMP_RNDN);
      for (j=1; j<=i; j++)
	mpfr_div_ui (P[i], P[i], j, GMP_RNDN);
      if (i%4==3)
	mpfr_mul_si (P[i], P[i], -1, GMP_RNDN);
    }
    mpfr_mul (P[i], P[i], NN, GMP_RNDN);
    mpfr_set_z (tmp, T, GMP_RNDN);
    mpfr_div_z (tmp, tmp, N, GMP_RNDN);
    mpfr_pow_ui (tmp, tmp, i, GMP_RNDN);
    mpfr_mul (P[i], P[i], tmp, GMP_RNDN);
  }
#elif FUNCTION==COS
  mpfr_set_z (tmp, tm, GMP_RNDN);
  mpfr_div_z (tmp, tmp, N, GMP_RNDN);
  mpfr_cos (P[0], tmp, GMP_RNDN);
  mpfr_mul (P[0], P[0], NN, GMP_RNDN);
  for (i=1; i<=d; i++) {
    if (i%2==1){
      mpfr_sin (P[i], tmp, GMP_RNDN);
      for (j=1; j<=i; j++)
	mpfr_div_ui (P[i], P[i], j, GMP_RNDN);
      if (i%4==1)
	mpfr_mul_si (P[i], P[i], -1, GMP_RNDN);
    }
    else{
      mpfr_cos (P[i], tmp, GMP_RNDN);
      for (j=1; j<=i; j++)
	mpfr_div_ui (P[i], P[i], j, GMP_RNDN);
      if (i%4==2)
	mpfr_mul_si (P[i], P[i], -1, GMP_RNDN);
    }
    mpfr_mul (P[i], P[i], NN, GMP_RNDN);
    mpfr_set_z (tmp, T, GMP_RNDN);
    mpfr_div_z (tmp, tmp, N, GMP_RNDN);
    mpfr_pow_ui (tmp, tmp, i, GMP_RNDN);
    mpfr_mul (P[i], P[i], tmp, GMP_RNDN);
  }
#elif FUNCTION==ONEOVERX
  mpfr_set_z (tmp, tm, GMP_RNDN);
  mpfr_div_z (tmp, tmp, N, GMP_RNDN);
  mpfr_ui_div (tmp, 1, tmp, GMP_RNDN);
  mpfr_mul (P[0], tmp, NN, GMP_RNDN);
  for (i=1; i<=d; i++) {
    mpfr_mul_si (P[i], P[i-1], -1, GMP_RNDN);
    mpfr_mul_z (P[i], P[i], T, GMP_RNDN);
    mpfr_div_z (P[i], P[i], N, GMP_RNDN);
    mpfr_mul (P[i], P[i], tmp, GMP_RNDN);
  }
#elif FUNCTION==HALL
  mpfr_set_z (tmp, tm, GMP_RNDN);
  mpfr_div_z (tmp, tmp, N, GMP_RNDN);
  mpfr_pow_ui (P[0], tmp, 3, GMP_RNDN);
  mpfr_sqrt (P[0], P[0], GMP_RNDN);
  mpfr_ui_div (tmp, 1, tmp, GMP_RNDN);
  mpfr_mul (P[0], P[0], NN, GMP_RNDN);
  mpfr_mul (P[1], P[0], tmp, GMP_RNDN);
  mpfr_mul_ui (P[1], P[1], 3, GMP_RNDN);
  mpfr_div_ui (P[1], P[1], 2, GMP_RNDN);
  mpfr_mul_z (P[1], P[1], T, GMP_RNDN);
  mpfr_div_z (P[1], P[1], N, GMP_RNDN);
  for (i=2; i<=d; i++) {
    mpfr_mul_si (P[i], P[i-1], -1, GMP_RNDN);
    mpfr_mul_z (P[i], P[i], T, GMP_RNDN);
    mpfr_div_z (P[i], P[i], N, GMP_RNDN);
    mpfr_mul (P[i], P[i], tmp, GMP_RNDN);
    mpfr_div_ui (P[i], P[i], i, GMP_RNDN);
    mpfr_div_ui (P[i], P[i], 2, GMP_RNDN);
    mpfr_mul_si (P[i], P[i], 2*i-5, GMP_RNDN);
  }
#elif FUNCTION==SQROOT

  mpfr_set_z (tmp, tm, GMP_RNDN);
  mpfr_div_z (tmp, tmp, N, GMP_RNDN);
  mpfr_sqrt (P[0], tmp, GMP_RNDN);
  mpfr_ui_div (tmp, 1, tmp, GMP_RNDN);
  mpfr_mul (P[0], P[0], NN, GMP_RNDN);
  for (i=1; i<=d; i++) {
    mpfr_mul_si (P[i], P[i-1], -1, GMP_RNDN);
    mpfr_mul_z (P[i], P[i], T, GMP_RNDN);
    mpfr_div_z (P[i], P[i], N, GMP_RNDN);
    mpfr_mul (P[i], P[i], tmp, GMP_RNDN);
    mpfr_div_ui (P[i], P[i], i, GMP_RNDN);
    mpfr_div_ui (P[i], P[i], 2, GMP_RNDN);
    mpfr_mul_si (P[i], P[i], 2*i-3, GMP_RNDN);
  }
#elif FUNCTION==SQRTINVERT
  mpfr_set_z (tmp, tm, GMP_RNDN);
  mpfr_div_z (tmp, tmp, N, GMP_RNDN);
  mpfr_ui_div (tmp, 1, tmp, GMP_RNDN);
  mpfr_sqrt (P[0], tmp, GMP_RNDN);
  mpfr_mul (P[0], P[0], NN, GMP_RNDN);
  for (i=1; i<=d; i++) {
    mpfr_mul_si (P[i], P[i-1], -1, GMP_RNDN);
    mpfr_mul_z (P[i], P[i], T, GMP_RNDN);
    mpfr_div_z (P[i], P[i], N, GMP_RNDN);
    mpfr_mul (P[i], P[i], tmp, GMP_RNDN);
    mpfr_div_ui (P[i], P[i], i, GMP_RNDN);
    mpfr_div_ui (P[i], P[i], 2, GMP_RNDN);
    mpfr_mul_ui (P[i], P[i], 2*i-1, GMP_RNDN);
  }
#endif

  if (near==1) mpfr_sub (P[0], P[0], target, GMP_RNDN);


#ifdef DEB2
  printf ("P:=");
    for (i=0; i<=d; i++){
      mpfr_out_str (stdout, 10, 10, P[i], GMP_RNDN);
      printf("*x^%d + ",i);
    }
    printf("\n");
#endif
    
    /* Look for the solutions of the equation
       |P(x_0/T) cmod 1| < 1/C, 
       where  |x0| \leq T and C = floor (1/(1/M+eps)).*/


    success = slz (&Res, &Nb, P, C, T, d, alpha, tmp, &B, &G, &U, &W, 
		   npols, ncoefs, ztmp, Calpha, ztmp2, ok, PP, Ptmp);

#ifdef DEB2
    printf ("success is %d, Nb is %d\n", success, Nb);
      if (Nb>0) gmp_printf ("Res[0] is %Zd\n", Res[0]);
#endif
    
    if (success == -1) {
      /*in case it fails, we divide T by 2 and rerun the algorithm twice*/
      failures ++;


#ifdef AUTOMATIC      
      if ((good>0)&&((100.0*failures)/good > FAILURE_UPPERBOUND)){
	mpz_mul_ui (initial_T, initial_T, 9);
	mpz_div_ui (initial_T, initial_T, 10);
	
	mpfr_set_d (Mprime, 0.9, GMP_RNDU);
	mpfr_pow_ui (Mprime, Mprime, d+1, GMP_RNDU);
	mpfr_mul (initial_eps, initial_eps, Mprime, GMP_RNDU);
	
	mpfr_ui_div (Mprime, 1, M, GMP_RNDN);
	mpfr_add (Mprime, Mprime, initial_eps, GMP_RNDN);
	mpfr_ui_div (Mprime, 1, Mprime, GMP_RNDN);
	mpfr_div_ui (Mprime, Mprime, 2, GMP_RNDN);
	mpfr_get_z (initial_C, Mprime, GMP_RNDD);
	
      }
#endif


      mpz_div_ui (T, T, 2);

      if (mpz_cmp_ui (T, 0)==0)  {
	fprintf (stderr, "FAIL at x=");
	mpz_out_str (stderr, 10, t0);
	fprintf (stderr, "\n");
	exit (1);
      }

      mpfr_div_2exp (eps, eps, d+1, GMP_RNDU);

      mpfr_ui_div (Mprime, 1, M, GMP_RNDN);
      mpfr_add (Mprime, Mprime, eps, GMP_RNDN);
      mpfr_ui_div (Mprime, 1, Mprime, GMP_RNDN);
      mpfr_div_ui (Mprime, Mprime, 2, GMP_RNDN);
      mpfr_get_z (C, Mprime, GMP_RNDD);
      mpz_pow_ui (Calpha, C, alpha);
    }
    else {
      /* update total_T */
      total_T += (2.0* (mpz_get_d (T))+1.0);          
      good++;
      
#ifdef AUTOMATIC      
      if ((good>0)&&((100.0*failures)/good < FAILURE_LOWERBOUND)){
	mpz_mul_ui (initial_T, initial_T, 10);
	mpz_div_ui (initial_T, initial_T, 9);

	mpfr_set_d (Mprime, 10.0/9.0, GMP_RNDU);
	mpfr_pow_ui (Mprime, Mprime, d+1, GMP_RNDU);
	mpfr_mul (initial_eps, initial_eps, Mprime, GMP_RNDU);
	
	mpfr_ui_div (Mprime, 1, M, GMP_RNDN);
	mpfr_add (Mprime, Mprime, initial_eps, GMP_RNDN);
	mpfr_ui_div (Mprime, 1, Mprime, GMP_RNDN);
	mpfr_div_ui (Mprime, Mprime, 2, GMP_RNDN);
	mpfr_get_z (initial_C, Mprime, GMP_RNDD);
      }
#endif

#ifdef DEB2
      printf ("total_T is %E\n", total_T);
      printf ("failures is %lld\n", failures);
#endif
      
      /* Remove several occurences of the same number */ 
      for (i=0; i<Nb; i++)
	for (j=i+1; j<Nb; j++){
	  if (mpz_cmp (Res[i], Res[j]) == 0){
	    for (k=j; k<Nb-1; k++)
	      mpz_set (Res[k], Res[k+1]);
	    Nb--;
	  }
	}


      /* check potential solutions and print solutions*/
      number += Check (Res, Nb, tm, N, NN, oneoverM, T, e_in);
      

      for (i=0; i<Nb; i++) mpz_clear (Res[i]);
      if (Nb>0) free(Res);

      mpz_add (t0, tm, T);
      mpz_add_ui (t0, t0, 1);
      
      /* If T is smaller than initial_T, then reset it */
      mpz_set (T, initial_T);
      mpfr_set (eps, initial_eps, GMP_RNDU);
      mpz_set (C, initial_C);
      mpz_pow_ui (Calpha, C, alpha);


#ifdef VBSE
    testtmp = 100.0*(mpz_get_d(t0) - t0init)/total_length;
    if ((testtmp > test_i) && (test_i != 100.0)){
      fprintf (stderr, "%d/100", (signed int) test_i);
      test_i += 10.0;
      fprintf (stderr, " x=");
      mpz_out_str (stderr, 10, t0);
      fprintf (stderr, "\n");
      fflush (stdout);
    }
#endif
    }
  }

#ifdef VBSE
  fprintf (stderr, "\nFailure rate is %lld pc.\n", (100*failures)/good);
  fprintf (stderr, "Number of calls to LLL: %lld. \n", failures+good);
  fprintf (stderr, "You got %d solution(s).\n", number);
  fprintf (stderr, "Done\n");
#ifdef AUTOMATIC
  mpfr_set_z (tmp, initial_T, GMP_RNDN);
  mpfr_log (tmp, tmp, GMP_RNDN);
  mpfr_set_d (eps, (double) BASIS, GMP_RNDN);
  mpfr_log (eps, eps, GMP_RNDN);
  mpfr_div (tmp, tmp, eps, GMP_RNDN);
  fprintf (stderr, "Final size of subintervals (log[%d](T)): %E.\n",
	   BASIS, mpfr_get_d (tmp, GMP_RNDN));
#endif
#endif

  for (i=0; i<=d; i++) mpfr_clear (P[i]); 
  free (P);

  mpfr_clear (tmp);
  mpfr_clear (eps);
  mpfr_clear (initial_eps);

  mpz_clear (initial_T);
  mpz_clear (t0);
  mpz_clear (t1);
  mpz_clear (tm);
  mpz_clear (N);
  mpfr_clear (NN);
  mpfr_clear (M);
  mpfr_clear (Mprime);
  mpfr_clear (oneoverM);
  mpz_clear (C);
  mpz_clear (initial_C);
  mpz_clear (T);
  mpz_clear (ztmp);
  mpz_clear (ztmp2);
  mpz_clear (Calpha);
  mpfr_clear (target);

  clear_matrix (&B, npols, ncoefs);
  clear_matrix (&W, npols, ncoefs);
  clear_matrix (&G, G.NumRows, G.NumCols);
  clear_matrix (&U, U.NumRows, U.NumCols);

  clear_vec (PP, d+1);
  clear_vec (Ptmp, ncoefs);
  free (ok);


  return 0;
}


